
/********************************************************************
 *
 * FILE:
 *	flanger.c
 *
 * DESCRIPTION:
 *	Attaches to the mixengined via the SLab APIs, and modifies data stream
 *	by recording the data into a local buffer, and feeding the local buffer
 *	back into the original stream with a short time delay. The delay is 
 *	modulated by sample insertion and extraction to create a signal phase.
 *
 * TOC:
 *	main()
 *	flange()
 *
 *******************************************************************/

#include "slabEffects.h"

/*
 * For historic reasons, the descriptor for bus attachments is int *.
 */
int *devPtr;

/*
 * Used by the user interface in the effects rack.
 */
#define EFFECT_NAME "flanger"
#define SHORT_NAME "Flange"

/*
 * These are dependant on the effect - parameters etc. This flanger implements
 * a single delay line, with a MAX delay, at 44100 samples, of about 200ms.
 * This HISTORY size may be excessive for simple flanging. Depth is limited
 * to about 1000 samples in any case, or just 1/50th of a second.
 */
#define HISTORY_SIZE 8192
int history[HISTORY_SIZE];
int indexIn = 0;
int indexOut = 0;
int	spread = 0;

int speed, breadth;
int Speed = 500, Breadth = 4096, Min = 1, currentBreadth = 0;

int changeBreadth = 1, countBreadth = 0;

int checkParams(), sampleRate;

/*
 * When the user interface attaches an effect, it passes two parameters, no
 * switches. This is very rudementary, but will be extended:
 *
 * argv[1] is the controlbuffer segID, argv[2] should be the bus requested
 */
main(argc, argv)
int argc;
char *argv[];
{
	int *busData, count, i;

	if (argc < 3)
	{
		printf("%s: parameter problem\n", argv[0]);
		exit(-1);
	}

	bzero(history, HISTORY_SIZE * sizeof(int));
	/*
	 * seAttach() will register with the mixengine, and link up the shared
	 * memory control and data structures. It effectively reserves a bus.
	 * The SLAB_EFFECTS_MONO really refers to the first bus attachment. We
	 * can attach to a second one with the SLAB_EFFECTS_STEREO flag. It is
	 * only used as a switch to use a slightly different link up protocol.
	 */
	if ((devPtr = seAttach(atoi(argv[1]), atoi(argv[2]),
		SLAB_EFFECTS_MONO)) ==
		(int *) NULL)
	{
		printf("%s: could not attach to requested bus\n", argv[0]);
		exit(-2);
	}

	/*
	 * At this point we can use seAddCtrl() to request up to 8 parameters
	 * which will be shown in the SLab effects user interface. The chorus algo
	 * requires speed, breadth. The calls return an int which is
	 * used later when requesting paramter values.
	 *
	 * Interface only supports the SLab_Potmeter controller type now.
	 */
	speed = seAddCtrl(devPtr, "Speed", SLab_Potmeter);
	breadth = seAddCtrl(devPtr, "Breadth", SLab_Potmeter);

	/*
	 * seInit() returns our fixed buffer size. We always use a fixed buffer as
	 * an internal data representation. No exceptions. Data is manipulated in
	 * this fixed buffer, and then the buffer (pointer) is returned to the bus.
	 *
	 * The buffer size is configurable in the GUI, to allow some memory tuning.
	 * For this reason we need to be able to find out the current value.
	 */
	sampleRate = seGetOpts(devPtr, SLAB_SAMPLE_RATE);
	checkParams(devPtr, 0);
	checkParams(devPtr, 1);
	count = seInit(devPtr, EFFECT_NAME, SHORT_NAME, checkParams);

	/*
	 * This is a forever loop.....
	 */
	while ((busData = seRead(devPtr)) != (int *) -1) {
		flange(busData, count);
		seWrite(devPtr, busData);
	}

	/*
	 * seDetach() will deregister with the mixengine, and remove links to the
	 * shared memory control and data structures.
	 */
	if (seDetach(devPtr) < 0)
	{
		printf("%s: could not detach from requested bus\n", argv[0]);
		exit(-3);
	}

	printf("Exiting\n");

	exit(0);
}

checkParams(busData, param)
register int *busData, param;
{
	char text[32];
	/*
	 * Load any new control values. The seGetCtrlVal() is actually quite
	 * efficient, it just returns a value set in the shared memory control
	 * buffer. It could be a macro, which would be faster, but would require
	 * knowledge of the controlBuffer structures which we are trying to hide.
	 */
	if (param == speed) {
		Speed = (256 - seGetCtrlVal(devPtr, speed)) << 2;

		sprintf(text, "Speed: %2.2f Hz", (float) sampleRate / (Speed*Breadth));
		seSetCtrlDisp(devPtr, speed, text);

		return;
	}
	Breadth = seGetCtrlVal(devPtr, breadth) + 1;

	sprintf(text, "Breadth: %3.1f", (float) (Breadth * 1000000)/sampleRate);
	seSetCtrlDisp(devPtr, breadth, text);
}


flange(busData, count)
register int *busData, count;
{
	for (; count > 0; count--)
	{
		/* 
		 * Output feed is at fixed rate. Input feed will insert/extract 
		 * samples to induce phasing. Breadth is limited to 256 samples at the
		 * moment, which is about 50ms at 44100Hz. This is MORE than enough.
		 */
		if (++countBreadth >= Speed)
		{
			countBreadth = 0;

			if ((currentBreadth += changeBreadth) >= Breadth)
				changeBreadth = -1;
			if (currentBreadth <= Min)
				changeBreadth = 1;

			if (changeBreadth > 0)
			{
				/*
				 * We are expanding our window, insert a sample.
				 */
				history[indexIn] = *busData;
				if (++indexIn >= HISTORY_SIZE)
					indexIn = 0;
				history[indexIn] = *busData;
				if (++indexIn >= HISTORY_SIZE)
					indexIn = 0;
			} /* else skip this sample = doNothing */
		} else {
			history[indexIn] = *busData;
			if (++indexIn >= HISTORY_SIZE)
				indexIn = 0;
		}

		*(busData++) = history[indexOut];

		if (++indexOut >= HISTORY_SIZE)
			indexOut = 0;
	}
}

